#
# re_graph.pl -- Graph a regular expression
#
=pod

=head1 NAME

re_graph.pl - Graph regular expression

=head1 SYNOPSIS

B<re_graph.pl> 
[B<-d>] 
[B<-f> I<re_file>]
[B<-P>] 
[I<regular expression>] 
[I<regular expression>] 
[I<...>] 

=head1 DESCRIPTION

The I<re_graph.pl> program graphs regular expressions.  The guts of the 
regular expression engine is a simple state machine.  The various states
and operations in the regular expression parser can be displayed using a 
surprisingly simple diagram.

A few notes on what you are looking at: 

The nodes B<Start> and B<Stop> denote the beginning and end of the regular
expression.  

The solid squares denote atoms.   Lines indicate the next state.
When a line splits, the state machine will take the top line first.
If it's path is blocked it will backup and take the next lower line.
This is repeated until it finds a path to the end or all paths are exhausted.

Brown boxes indicate a grouping operation, i.e. ().  

Green boxes indicate a zero with test.  The state machine will perform the 
test inside the box before moving ahead.

For more information, see the tutorial below.

=head1 OPTIONS

=over 4

=item B<-d>

Turn on debugging.  The debugging output is printed on the console as regular
expressions are compiled.

=item B<-f> I<file>

Specify a file containing a list of regular expressions, one per line, to be
put in the regular expression list.

=item B<-P> 

Enable the I<Print> button.  This button doesn't work correctly right now,
but you can play with it if you want.  When I<Print> is pressed, the results
are written into a Postscript file named I<tmp.ps>.

=back

Note: If no regular expressions are specified, a list consisting of the
items in the tutorial is loaded.

=head1 GUI CONTROLS

=over 4 

=item B<Next>

Displays the next regular expression in the list.

=item B<Previous>

Displays the previous regular expression in the list.

=item I<Regular Expression Blank>

This blank contains the regular expression being graphed.

=item B<List>

Pops up a window containing a list of expressions.  You can select an 
expression from this list and press B<OK> to graph it.  You can also
input a regular expression in the blank at the bottom of the window 
and press B<New> to add it to the list.

=item B<Exit>

Exits the program.

=item B<Print>

If it worked, it would print the current graph.  But since it is broken
it won't even show up unless you put the B<-P> switch on the command line.

When pressed it creates a file called I<tmp.ps> which should contain a 
postscript version of your graph.   (It has a few problems that prevent the
output from being what you want.)

=back

=head1 TUTORIAL

This tutorial shows what happens when a set of sample regular expressions
are graphed.  This set of regular expressions closely follows the
Chapter 4 of "Perl for C Programmers" by Steve Oualline.

The set of regular expressions used for this tutorial is:

    test
    ^Test
    ^ *#
    ^[ \t]*#
    ^\s*#
    ([^#]*)(#.*)
    a|b
    ^(([^#"]*|".*")*)(#.*)
    ^((?:[^#"]*|".*?")*)(#.*)
    ^((?:[^#"]*|".*?(?<!\\)")*)(#.*)

Let's take a look at the graphs produced by these expressions.

=over 4

=item B</test/>

This is a very simple expression.  It matches "test" anywhere on the line.
If you look at the graph of this expression, it consists of three nodes "Start",
"EXACT <test>" and "END".   

The "Start" node indicates the start of the regular expression.  

The "EXACT <test>" node tells the engine that the text must match the
text "test" exactly.   

The "END" node indicates the end of the regular expression.  If you reach
the "END" node, a successful match was found.

Flow is a straight line from "Start", through the "EXACT" check, to end.

=item B</^Test/>

A new item was added with this expression, an anchor.  It's named BOL 
(Beginning of line) and shows up as an additional node.

=item B</^ *#/>

Now we start having fun.  This expression matches anything that consists
of a start of line (^), a bunch of spaces ( *), and a sharp (#).

The way the state machine works it that it starts at "Start" and works
it's way through the nodes.  You'll notice that between "BOL" and
"EXACT < >" there's a fork in the road.  

The state machine will take the top branch if possible.  So if the next
character is a space, the system will take the top branch and match the
"EXACT < >" node.  If not, the bottom branch is taken and we wind up
at the "EXACT <#>" node. 

If there's no path to the "END", then we don't have a match.

=item B</^[ \t]*#/>

This is the same as the previous example, except the space was replaced
by a character set.  We call the set "space and tab".  The system translates
this into "\11\40".  It's the same thing, suitable obfuscated for computer
work.

=item B</^\s*#/>

Again, the middle as been replace by another token.  In this case it's 
the SPACE token which matches any whitespace.

=item B</([^#]*)(#.*)/>

This expression introduces us to the grouping operators.  They show as the 
big brown boxes.

The other change is that we use the expression [^#], which matches anything
except a hash mark.  Perl changes this to a "ANYOF" clauses which matches
all characters except the single one we don't want.  

Note: This ANYOF node overflows the size of the box.  This is a know bug.

=item B</a|b/>

Now we introduce the concept of a selection of two different atoms.  Note that
the branch arrows are drawn smaller to make them stand out.

=item B</^(([^#"]*|".*")*)(#.*)/>

See the book for what this regular expression tries to match.

This expression adds nested grouping, and some additional stuff that we've seen
before.

=item B</^((?:[^#"]*|".*?")*)(#.*)/>

This is like the previous example, except what was the $2 grouping has been
replaced by the "Group no $" operator (?:...).  Notice that the line around
the second group has disappeared and what was $3 is now $2.

(In future versions of this graphing tool we will graph the invisible group
operator.  We just did figure out how to do it yet.)

Also notice the use of the "*?" operator.  Remember when going through
the nodes, when a branch is encountered, the system will try and take
the top one first.


=item B</^((?:[^#"]*|".*?(?<!\\)")*)(#.*)/>

The grand finale.  One new type of node has been introduced: (?<!\\).  This is 
the negative look-behind.  It's the red box on the screen.  When the state machine
sees this, it matches the text behind the current location marker against the
indicated text and if it fails then a match against the next node is possible.
(Boy does this not translate into English well.)

Basically the clause in question looks for a double quoted string ("xxx"), but
will ignore a double quote it's escaped ("xxx\"yyy").

=back

=head1 BUGS / LIMITATIONS

This will not graph all the regular expressions.  Some of the more advanced
features of the engine are just not handled.

We currently "graph" the "group, no $1" (?:..) operator by displaying nothing.
A box should be put around the expression.

The boxes drawn by the program are a fixed with not related to the size of 
the text they contain.  Text can easily overflow the box.

The system is UNIX/Linux specific.  This is caused by only one small
section of code should anyone want to port this to a braindamaged
operating system.

Better use of color can be made.   Specifically all the nodes do not
have to be green.  Come to think of it they call don't have to be
rectangles either.

Sometimes the lines connecting one section to another take some strange
twists.

Horizontal scrolling of the regular expression could be better.

=head1 FUTURE WORK

I would like to connect this to the execution of the regular expression
engine so we can see the thing in action.

=head1 LICENSE

Licensed under the GPL.  (See the end of the source file for a copy).

=head1 AUTHOR

Steve Oualline (oualline@www.oualline.com)

=cut
use strict;
use warnings;

use Tk;
use Tk::LabFrame;
use Tk::FileSelect;
use Tk::Font;
use Tk::ROText;

use IO::Handle;
use English;

use constant X_NODE_SIZE => 60;   # Size of a node (X Space)
use constant Y_NODE_SIZE => 40;   # Size of a node (Y Space)

use constant X_TEXT_OFFSET => 3;  # Space text over this far
use constant Y_TEXT_OFFSET => 3;  # Space text over this far

use constant X_MARGIN => 50;      # Space between nodes (X)
use constant Y_MARGIN => 10;      # Vertical spacing

use Getopt::Std;

use vars qw/$opt_d $opt_f $opt_P/;

STDOUT->autoflush(1);

my $tk_scrolled_canvas;	# Canvas with scrollbars
my $tk_canvas;		# The canvas we are using for regular expression

my $tk_font;            # Font we use for * loops
my $tk_small_font;      # Font we use () captions

my $tk_re_list_window;  # Window with list of regular expressions
my $tk_list;            # List of regular expressions

my $tk_mw;		# Main window
my $tk_current_re;      # The text box containing the current RE

# Configuration items
my $x_margin = 16;	# Space between items
my $y_margin = 16;	# Space between items

#
# Fields
#       node    -- Node number
#       type    -- Node type (from re debug)
#       arg     -- Argument (optional)
#       next    -- Next node
#
my @re_debug;           # Regular expression debugging information
#
# Fields
#       x_size    - Size of the node in X
#       y_size    - Size of the node in Y
#       x_loc     - X Location of the node
#       y_loc     - Y Location of the node
#       node      - Reference to the node in @re_debug
#       child     - Array of child nodes for this node
#
my @format_re;          # Formatted version of the regular expression

my $current_re;         # Re we are displaying now
my @re_list;            # Regular expression list

my $re_to_add = "";     # Re we are adding

sub draw_node_array($);
sub size_array(\@);
sub layout_array($$$\@);
############################################################
# The "PLUS" node
#
#
#     0  1  2    1p 2p  3p (p = +size of child)
#     v  v  v L3 v  v   v
#     .  ---------  .   .
#     . /.   .   .\ .   .
#     ./ .   .   . \    .
# a2  <  .   .a1 .  >   .
#     .\ .   .   . /.   .
#     . \+-------+/     .
#  L1--->| child |----->+ L2
#     .  +-------+  .   .
#
# Bounding boxes
#
#       a1 / 270  - 180 / (ap, y-2a), (a3p, y)
#       a2 /  90  - 180 / (a0, y-2a), (a2, y-2a)
#
#       L1 (a3, y+2a) (a3p, y+2a)
############################################################

#------------------------------------------------------------
# size_plus -- Compute the size of a star type node
#------------------------------------------------------------
sub size_plus($)
{
    my $node = shift;   # Node we want layout information for

    # Compute the size of the children
    my ($x_size, $y_size) = size_array(@{$node->{children}});

    # Arc size is based on the Y dimension of the children
    $node->{arc_size} = int($y_size/4) + 10;
    #TODO: Make the 10 constant (it's the space over/below)

    $node->{child_x} = $x_size - X_MARGIN;

    $node->{x_size} = $node->{child_x} + $node->{arc_size} * 2 + X_MARGIN;
    $node->{y_size} = $y_size + $node->{arc_size} * 2;
}
#------------------------------------------------------------
# Draw the plus type node
#------------------------------------------------------------
sub draw_plus($)
{
    my $cur_node = shift;       # The node we are drawing

    layout_array(
        $cur_node->{x_loc} + $cur_node->{arc_size} * 1,
        $cur_node->{y_loc},
        $cur_node->{y_size},
        @{$cur_node->{children}});

    draw_node_array($cur_node->{children});

    # The place we start drawing from (X)
    my $from_x = $cur_node->{x_loc};

    # The current middle of the item (Y)
    my $y = $cur_node->{y_loc} + ($cur_node->{y_size}/2);

    my $arc_size = $cur_node->{arc_size};       # Size of an arc

    # Size of the child
    my $child_x = $cur_node->{child_x};

    # Debugging
    if (0) {
        for (my $debug_x = 0; $debug_x < 5; $debug_x++) {
            $tk_canvas->createLine(
                    $from_x + $arc_size * $debug_x, $y - $arc_size*2,
                    $from_x + $arc_size * $debug_x, $y + $arc_size*2,
                    );
        }

        for (my $debug_x = 3; $debug_x < 7; $debug_x++) {
            $tk_canvas->createLine(
                    $from_x + $child_x + $arc_size * $debug_x, $y - $arc_size*2,
                    $from_x + $child_x + $arc_size * $debug_x, $y + $arc_size*2,
                    -fill => "Green",
                    );
        }
    }

    my $flip = 1;       # Flipping factor
    if ($cur_node->{min_flag}) {
        $flip = -1;
    }

    # First arc (a1)
    $tk_canvas->createArc(
            $from_x + $child_x + $arc_size * 0, $y - $arc_size * 2 * $flip,
            $from_x + $child_x + $arc_size * 2, $y,
            -start => ($flip == 1) ? 270 : 90,
            -extent => $flip * 180,
            -outline => "Black",
            -style => "arc",
            -width => 3);

    # Second arc (a2)
    $tk_canvas->createArc(
            $from_x + $arc_size * 0, $y - $arc_size * 2 * $flip,
            $from_x + $arc_size * 2, $y,
            -start => ($flip == 1) ? 90 : 270,
            -extent => 180 * $flip,
            -outline => "Black",
            -style => "arc",
            -width => 3);

    # Draw (L1)
    $tk_canvas->createLine(
            $from_x, $y,
            $from_x + $arc_size * 1, $y,
            -arrow => "last",
            -tags => "arrow",
            -width => 3,
            -fill => "Black"
    );

    # Draw (L2)
    $tk_canvas->createLine(
            $from_x + $child_x + $arc_size * 1, $y,
            $from_x + $child_x + $arc_size * 2, $y,
            -arrow => "last",
            -tags => "arrow",
            -width => 3,
            -fill => "Black"
    );

    # Draw (L3)
    $tk_canvas->createLine(
            $from_x + $child_x + $arc_size * 1, $y - $arc_size * 2,
            $from_x + $arc_size * 1, $y - $arc_size * 2,
            -arrow => "last",
            -tags => "arrow",
            -width => 3,
            -fill => "Black"
    );


    $tk_canvas->createText(
            $from_x + $child_x + $arc_size * 2, $y - $arc_size * 2,
            -text => ($cur_node->{min_flag}) ? "+?" : "+",
            -fill => "Black",
            -font => $tk_font);

    $cur_node->{left_x} = $from_x;
    $cur_node->{left_y} = $y;

    $cur_node->{right_x} = $from_x + $cur_node->{child_x} +
                            $cur_node->{arc_size} * 2;
    $cur_node->{right_y} = $y;
}
############################################################
# The "STAR" node
#
#
#     0  1  2    3       3p 4p  5p (p = +size of child)
#     v  v  v    v   L2  v  v   v
#     .  -----------------  .   .
#     . /.  .    .       .\ .   .
#     ./ .  .    .       . \    .
# a6  <  .  .    .    a5 .  >   .
#     .\ .  .    .       . /.   .
#     . \.  . .  +-------+/     .
#  L3----------->| child |- .   +
#     .  .\ . j  +-------+  .a4/.
#     .  . \a1   .       .  . / .
#     .  .  \    .       .  ./  .
#     .  .  |    .       .  |   .
#     .  .  .\   .       . /   .
#     .  .  a2\  .       ./a3  .
#     .  .  .  \---------
#           ^    ^    L1
#           2    3
#
# Bounding boxes
#       a1 /   0  - 90  / (a0,  y), (a2,  y+2a)
#       a2 /  90  - 0   / (a2,  y), (a4,  y+2a)
#       a3 /   0  - -90 / (a2p, y), (a4p, y+2a)
#       a4 / -90  - 0   / (a4p, y), (a6p, y+2a)
#
#       a5 / 270  - 180 / (a2p, y-2a), (a4p, y)
#       a6 /  90  - 180 / (a0,  y-2a), (a2, y-2a)
#
#       L1 (a3, y+2a) (a3p, y+2a)
############################################################

#------------------------------------------------------------
# size_star -- Compute the size of a star type node
#------------------------------------------------------------
sub size_star($)
{
    my $node = shift;   # Node we want layout information for

    # Compute the size of the children
    my ($x_size, $y_size) = size_array(@{$node->{children}});

    # Arc size is based on the Y dimension of the children
    $node->{arc_size} = int($y_size/4) + 10;
    #TODO: Make the 10 constant (it's the space over/below)

    $node->{child_x} = $x_size - X_MARGIN;

    $node->{x_size} = $node->{child_x} + $node->{arc_size} * 5 + X_MARGIN;
    $node->{y_size} = $y_size + $node->{arc_size} * 2 + Y_MARGIN;


}
#------------------------------------------------------------
# Draw the star type node
#------------------------------------------------------------
sub draw_star($)
{
    my $cur_node = shift;       # The node we are drawing

    layout_array(
        $cur_node->{x_loc} + $cur_node->{arc_size} * 3,
        $cur_node->{y_loc},
        $cur_node->{y_size},
        @{$cur_node->{children}});

    draw_node_array($cur_node->{children});

    # The place we start drawing from (X)
    my $from_x = $cur_node->{x_loc};

    # The current middle of the item (Y)
    my $y = $cur_node->{y_loc} + ($cur_node->{y_size}/2);

    my $arc_size = $cur_node->{arc_size};       # Size of an arc

    # Size of the child
    my $child_x = $cur_node->{child_x};

    # Debugging
    if (0) {
        for (my $debug_x = 0; $debug_x < 5; $debug_x++) {
            $tk_canvas->createLine(
                    $from_x + $arc_size * $debug_x, $y - $arc_size*2,
                    $from_x + $arc_size * $debug_x, $y + $arc_size*2,
                    );
        }

        for (my $debug_x = 3; $debug_x < 7; $debug_x++) {
            $tk_canvas->createLine(
                    $from_x + $child_x + $arc_size * $debug_x, $y - $arc_size*2,
                    $from_x + $child_x + $arc_size * $debug_x, $y + $arc_size*2,
                    -fill => "Green",
                    );
        }
    }

    my $flip = 1;       # Flipping factor
    if ($cur_node->{min_flag}) {
        $flip = -1;
    }

    # First arc (a1)
    $tk_canvas->createArc(
            $from_x + $arc_size * 0, $y,
            $from_x + $arc_size * 2, $y + $arc_size * 2 * $flip,
            -start => 0,
            -extent => 90 * $flip,
            -outline => "Black",
            -style => "arc",
            -width => 3);

    # Second arc (a2)
    $tk_canvas->createArc(
            $from_x + $arc_size * 2, $y,
            $from_x + $arc_size * 4, $y + $arc_size * 2 * $flip,
            -start => 180,
            -extent => 90 * $flip,
            -outline => "Black",
            -style => "arc",
            -width => 3);

    # Third arc (a3)
    $tk_canvas->createArc(
            $from_x + $child_x + $arc_size * 2, $y,
            $from_x + $child_x + $arc_size * 4, $y + $arc_size * 2 * $flip,
            -start => 0,
            -extent => -90 * $flip,
            -outline => "Black",
            -style => "arc",
            -width => 3);

    # Fourth arc (a4)
    $tk_canvas->createArc(
            $from_x + $child_x + $arc_size * 4, $y,
            $from_x + $child_x + $arc_size * 6, $y + $arc_size * 2 * $flip,
            -start => ($flip == 1) ? 90 : 180,
            -extent => 90,
            -outline => "Black",
            -style => "arc",
            -width => 3);

    # Fifth arc (a5)
    $tk_canvas->createArc(
            $from_x + $child_x + $arc_size * 2, $y - $arc_size * 2 * $flip,
            $from_x + $child_x + $arc_size * 4, $y,
            -start => ($flip == 1) ? 270 : 90,
            -extent => $flip * 180,
            -outline => "Black",
            -style => "arc",
            -width => 3);

    # Sixth arc (a6)
    $tk_canvas->createArc(
            $from_x + $arc_size * 0, $y - $arc_size * 2 * $flip,
            $from_x + $arc_size * 2, $y,
            -start => ($flip == 1) ? 90 : 270,
            -extent => 180 * $flip,
            -outline => "Black",
            -style => "arc",
            -width => 3);

    # L1
    $tk_canvas->createLine(
            $from_x + $arc_size * 3, $y + $arc_size * 2 * $flip,
            $from_x + $arc_size * 3 + $child_x, $y + $arc_size * 2 * $flip,
            -arrow => "last",
            -tags => "arrow",
            -width => 3,
            -fill => 'Black');

    # L2
    $tk_canvas->createLine(
            $from_x + $arc_size * 3 + $child_x, $y - $arc_size * 2 * $flip,
            $from_x + $arc_size * 1, $y - $arc_size * 2 * $flip,
            -arrow => "last",
            -tags => "arrow",
            -width => 3);

    # Draw (L3)
    $tk_canvas->createLine(
            $from_x, $y,
            $from_x + $arc_size * 3, $y,
            -arrow => "last",
            -tags => "arrow",
            -width => 3,
            -fill => "Black"
    );


    $tk_canvas->createText(
            $from_x + $child_x + $arc_size * 4, $y - $arc_size * 2,
            -text => ($cur_node->{min_flag}) ? "*?" : "*",
            -fill => "Black",
            -font => $tk_font);


    $cur_node->{left_x} = $from_x;
    $cur_node->{left_y} = $y;

    $cur_node->{right_x} = $from_x + $cur_node->{child_x} +
                            $cur_node->{arc_size} * 5;
    $cur_node->{right_y} = $y;
}

############################################################
# Branch nodes
############################################################
use constant X_BRANCH_MARGIN => 20;     # Space between branches (x)
use constant Y_BRANCH_MARGIN => 20;     # Space between branches (y)
#-----------------------------------------------------------
# layout a branch node
#-----------------------------------------------------------
sub size_branch($)
{
    my $node = shift;   # Node we want layout information for

    my $x_size = 0;     # Current X size
    my $y_size = 0;     # Current Y size

    foreach my $cur_choice (@{$node->{choices}}) {
        # The size of the current choice
        my ($x_choice, $y_choice) = size_array(@{$cur_choice});

        if ($x_size < $x_choice) {
            $x_size = $x_choice;
        }
        if ($y_size != 0) {
            $y_size += Y_BRANCH_MARGIN;
        }
        $cur_choice->[0]->{row_y_size} = $y_choice;
        $y_size += $y_choice;
    }
    $x_size += 2 * X_BRANCH_MARGIN + X_MARGIN;
    $node->{x_size} = $x_size;
    $node->{y_size} = $y_size;
}
#-----------------------------------------------------------
# draw_branch -- Draw a branch structure
#-----------------------------------------------------------
sub draw_branch($)
{
    my $cur_node = shift;   # Node we want layout information for

    # Location where we draw the branches
    my $x_loc = $cur_node->{x_loc} + X_BRANCH_MARGIN;
    my $y_loc = $cur_node->{y_loc};

    foreach my $cur_child (@{$cur_node->{choices}}) {
        layout_array(
            $x_loc + X_BRANCH_MARGIN,
            $y_loc,
            $cur_child->[0]->{row_y_size},
            @{$cur_child});
        $y_loc += $cur_child->[0]->{row_y_size} + Y_BRANCH_MARGIN;
        draw_node_array($cur_child);
    }

    my $max_x = 0;      # Largest right x of any node

    foreach my $cur_child (@{$cur_node->{choices}}) {
        # Last node on the string of children
        my $last_node = $cur_child->[$#{$cur_child}];

        if ($last_node->{right_x} > $max_x) {
            $max_x = $last_node->{right_x};
        }
    }
    foreach my $cur_child (@{$cur_node->{choices}}) {
        # Last node on the string of children
        my $last_node = $cur_child->[$#{$cur_child}];

        if ($last_node->{right_x} < $max_x) {
            $tk_canvas->createLine(
                    $last_node->{right_x}, $last_node->{right_y},
                    $max_x, $last_node->{right_y},
                    -arrow => "last",
                    -tags => "arrow",
                    -width => 3,
                    -fill => "Black");
        }
    }

    my $left_x = $cur_node->{x_loc};
    my $right_x = $cur_node->{x_loc} + $cur_node->{x_size} - X_MARGIN;
    my $y = $cur_node->{y_loc} + ($cur_node->{y_size} / 2);

    foreach my $cur_child (@{$cur_node->{choices}}) {
        # Create a branch line to the item in the list of nodes
        $tk_canvas->createLine(
                $left_x, $y,
                $cur_child->[0]->{left_x}, $cur_child->[0]->{left_y},
                -arrow => "last",
                -tags => "arrow",
                -width => 1,
                -fill => "Black");

        # The last node on the list
        my $last_child = $cur_child->[$#$cur_child];
        # Line from the last node to the collection point
        $tk_canvas->createLine(
                $max_x, $last_child->{right_y},
                $right_x, $y,
                -arrow => "last",
                -tags => "arrow",
                -width => 1,
                -fill => "Black");
    }

    $cur_node->{left_x} = $left_x;
    $cur_node->{left_y} = $y;

    $cur_node->{right_x} = $right_x;
    $cur_node->{right_y} = $y;
}
############################################################
# Functions to compute the size of various nodes
############################################################
#-----------------------------------------------------------
# layout the start node
#-----------------------------------------------------------
sub size_start($)
{
    my $node = shift;   # Node we want layout information for
    $node->{x_size} = X_NODE_SIZE + X_MARGIN;
    $node->{y_size} = Y_NODE_SIZE;
}
#-----------------------------------------------------------
# layout the end node
#-----------------------------------------------------------
sub size_end($)
{
    my $node = shift;   # Node we want layout information for
    $node->{x_size} = X_NODE_SIZE;
    $node->{y_size} = Y_NODE_SIZE;
}
#-----------------------------------------------------------
# layout the "EXACT" node  (EXACT + text)
#-----------------------------------------------------------
sub size_exact($)
{
    my $node = shift;   # Node we want layout information for
    $node->{x_size} = X_NODE_SIZE + X_MARGIN;
    $node->{y_size} = Y_NODE_SIZE;
}
#-----------------------------------------------------------
# OPEN  the open (
#-----------------------------------------------------------
use constant BOX_MARGIN => 50;  # Size of the box around a group
use constant BOX_FONT_SIZE => 15;       # Height of the font used to label boxes

sub size_open($)
{
    my $node = shift;   # The node we want to size

    # Compute the size of the children
    my ($x_size, $y_size) = size_array(@{$node->{children}});

    # We add X_MARGIN because we must for all nodes
    # We subtract X_MARGIN because one too many is added in our children
    # Result is nothing
    $node->{x_size} = $x_size + BOX_MARGIN;
    $node->{y_size} = $y_size + BOX_MARGIN + BOX_FONT_SIZE;
}


my %compute_size = (
    "ANYOF" => \&size_exact,
    "BOL" => \&size_exact,
    "SPACE" => \&size_exact,
    "BRANCH"=> \&size_branch,
    "END"   => \&size_end,
    "EOL" => \&size_exact,
    "EXACT" => \&size_exact,
    "IFMATCH"  => \&size_open,
    "OPEN"  => \&size_open,
    "PLUS"  => \&size_plus,
    "REF"   => \&size_exact,
    "REG_ANY" => \&size_exact,
    "STAR"  => \&size_star,
    "Start" => \&size_start,
    "UNLESSM"  => \&size_open
);

############################################################
# draw functions
############################################################
sub draw_start_end($)
{
    my $cur_node = shift;
    my $node_number = $cur_node->{node}->{node};

    $cur_node->{text} = $tk_canvas->createText(
            $cur_node->{x_loc} + X_TEXT_OFFSET,
            $cur_node->{y_loc} + Y_TEXT_OFFSET,

            -text => $cur_node->{node}->{type},
            -fill => "Black",
            -anchor => "nw",
            -tags => ["text$node_number"]);

    $cur_node->{rectangle} = $tk_canvas->createRectangle(
            $cur_node->{x_loc}, $cur_node->{y_loc},
            $cur_node->{x_loc} + X_NODE_SIZE,
            $cur_node->{y_loc} + Y_NODE_SIZE,
            -fill => "Green",
            -tags => ["rect$node_number"]);
    $tk_canvas->raise("text$node_number", "rect$node_number");

    my @bbox = $tk_canvas->bbox("rect$node_number");

    $cur_node->{left_x} = $bbox[0];
    $cur_node->{left_y} = ($bbox[1] + $bbox[3]) / 2;

    $cur_node->{right_x} = $bbox[2];
    $cur_node->{right_y} = ($bbox[1] + $bbox[3]) / 2;
}

#-----------------------------------------------------------
# draw_exact($node) -- Draw a "EXACT" re node
#-----------------------------------------------------------
sub draw_exact($)
{
    my $cur_node = shift;       # The node
    my $node_number = $cur_node->{node}->{node};

    $cur_node->{text} = $tk_canvas->createText(
            $cur_node->{x_loc} + X_TEXT_OFFSET,
            $cur_node->{y_loc} + Y_TEXT_OFFSET,

            -text => "$cur_node->{node}->{type}\n  $cur_node->{node}->{arg}",
            -fill => "Black",
            -anchor => "nw",
            -tags => ["text$node_number"]);

    $cur_node->{rectangle} = $tk_canvas->createRectangle(
            $cur_node->{x_loc}, $cur_node->{y_loc},
            $cur_node->{x_loc} + X_NODE_SIZE,
            $cur_node->{y_loc} + Y_NODE_SIZE,
            -fill => "Green",
            -tags => ["rect$node_number"]);

    $tk_canvas->raise("text$node_number", "rect$node_number");

    my @bbox = $tk_canvas->bbox("rect$node_number");

    $cur_node->{left_x} = $bbox[0];
    $cur_node->{left_y} = ($bbox[1] + $bbox[3]) / 2;

    $cur_node->{right_x} = $bbox[2];
    $cur_node->{right_y} = ($bbox[1] + $bbox[3]) / 2;
}
#-----------------------------------------------------------
# draw_ref($node) -- Draw a "REF" re node
#-----------------------------------------------------------
sub draw_ref($)
{
    my $cur_node = shift;       # The node
    my $node_number = $cur_node->{node}->{node};

    $cur_node->{text} = $tk_canvas->createText(
            $cur_node->{x_loc} + X_TEXT_OFFSET,
            $cur_node->{y_loc} + Y_TEXT_OFFSET,

            -text => "Back Reference:\n  $cur_node->{node}->{ref}",
            -fill => "Black",
            -anchor => "nw",
            -tags => ["text$node_number"]);

    $cur_node->{rectangle} = $tk_canvas->createRectangle(
            $cur_node->{x_loc}, $cur_node->{y_loc},
            $cur_node->{x_loc} + X_NODE_SIZE,
            $cur_node->{y_loc} + Y_NODE_SIZE,
            -fill => "LightGreen",
            -tags => ["rect$node_number"]);

    $tk_canvas->raise("text$node_number", "rect$node_number");

    my @bbox = $tk_canvas->bbox("rect$node_number");

    $cur_node->{left_x} = $bbox[0];
    $cur_node->{left_y} = ($bbox[1] + $bbox[3]) / 2;

    $cur_node->{right_x} = $bbox[2];
    $cur_node->{right_y} = ($bbox[1] + $bbox[3]) / 2;
}
#-----------------------------------------------------------
# draw the () stuff
#-----------------------------------------------------------
sub draw_open($$)
{
    my $cur_node = shift;       # The node

    $cur_node->{rectangle} = $tk_canvas->createRectangle(
            $cur_node->{x_loc}, $cur_node->{y_loc} + BOX_FONT_SIZE,
            $cur_node->{x_loc} + $cur_node->{x_size} - X_MARGIN,
            $cur_node->{y_loc} + $cur_node->{y_size},
            -outline => $cur_node->{color},
            -dash => ".",
            -width => 3);

    $tk_canvas->createText(
            $cur_node->{x_loc}, $cur_node->{y_loc},
            -text => $cur_node->{text},
            -fill => $cur_node->{color},
            -anchor => "w",
            -font => $tk_small_font);

    layout_array(
        $cur_node->{x_loc} + BOX_MARGIN/2,
        $cur_node->{y_loc} + BOX_MARGIN/2 + BOX_FONT_SIZE,
        $cur_node->{y_size} - BOX_MARGIN - BOX_FONT_SIZE,
        @{$cur_node->{children}});

    draw_node_array($cur_node->{children});


    $cur_node->{left_x} = $cur_node->{x_loc};
    $cur_node->{left_y} = $cur_node->{y_loc} + ($cur_node->{y_size} + BOX_FONT_SIZE)/2;

    $cur_node->{right_x} = $cur_node->{x_loc} + $cur_node->{x_size} - X_MARGIN;
    $cur_node->{right_y} = $cur_node->{left_y};

    # Child we are drawing arrows to / from
    my $child = $cur_node->{children}->[0];
    $tk_canvas->createLine(
            $cur_node->{left_x}, $cur_node->{left_y},
            $child->{left_x}, $child->{left_y},
            -arrow => "last",
            -tags => "arrow",
            -width => 3,
            -fill => "Black"
    );
    $child = $cur_node->{children}->[$#{$cur_node->{children}}];
    $tk_canvas->createLine(
            $child->{right_x}, $child->{right_y},
            $cur_node->{right_x}, $cur_node->{right_y},
            -arrow => "last",
            -tags => "arrow",
            -width => 3,
            -fill => "Black"
    );
}

my %draw_node = (
    "ANYOF" => \&draw_exact,
    "BOL"   => \&draw_start_end,
    "EOL"   => \&draw_start_end,
    "SPACE"   => \&draw_start_end,
    "BRANCH"=> \&draw_branch,
    "END"   => \&draw_start_end,
    "EXACT" => \&draw_exact,
    "IFMATCH"  => \&draw_open,
    "OPEN"  => \&draw_open,
    "PLUS"  => \&draw_plus,
    "REF"   => \&draw_ref,
    "REG_ANY" => \&draw_start_end,
    "STAR"  => \&draw_star,
    "Start" => \&draw_start_end,
    "UNLESSM"  => \&draw_open
);


########################################################################
# usage -- Tell the user how to use us
########################################################################
sub usage()
{
    print STDERR
	"Usage is $0 [-f <file>] [-P] [<regular-expression>]\n";
    exit (8);
}
########################################################
# parse_re -- Parse a regular expression and leave
#       the results in the array @re
########################################################
sub parse_re()
{
    my $quote_re = $current_re;
    $quote_re =~ s/\\/\\\\/g;
    my $cmd = <<EOF ;
perl 2>&1 <<SHELL_EOF
use re 'debug';
/$quote_re/;
SHELL_EOF
EOF

    # The raw debug output
    my @raw_debug = `$cmd`;

    if ($opt_d) {
        print @raw_debug;
    }

    if ($CHILD_ERROR != 0) {
    my $cmd = <<EOF ;
perl 2>&1 <<SHELL_EOF
use re 'debug';
/ERROR/;
SHELL_EOF
EOF
        @raw_debug = `$cmd`;
        if ($CHILD_ERROR != 0) {
            die("Could not run perl");
        }
    }

    @re_debug = ();     # Clear out old junk
    push(@re_debug, {
            node => 0,
            type => "Start",
            next => 1
            });
    foreach my $cur_line (@raw_debug) {
        if ($cur_line =~ /^Compiling/) {
            next;
        }
        if ($cur_line =~ /^\s*size/) {
            next;
        }
        #                 +++---------------------------------- Spaces
        #                 ||| +++------------------------------ Digits
        #                 |||+|||+----------------------------- Group $1
        #                 ||||||||                              (Node)
        #                 ||||||||
        #                 ||||||||+---------------------------- Colon
        #                 |||||||||+++------------------------- Spaces
        #                 ||||||||||||
        #                 |||||||||||| +++--------------------- Word chars
        #                 ||||||||||||+|||+-------------------- Group $2
        #                 |||||||||||||||||                       (Type)
        #                 |||||||||||||||||
        #                 |||||||||||||||||+++----------------- Spaces
        #                 ||||||||||||||||||||
        #                 |||||||||||||||||||| ++--------------- Any char str
        #                 ||||||||||||||||||||+||+-------------- Group $3
        #                 ||||||||||||||||||||||||               (arg)
        #                 ||||||||||||||||||||||||------------- Lit <>
        #                 ||||||||||||||||||||||||
        #                 ||||||||||||||||||||||||+++---------- Spaces
        #                 |||||||||||||||||||||||||||
        #                 |||||||||||||||||||||||||||   ++----- Any char str
        #                 |||||||||||||||||||||||||||++ || ++-- Lit ()
        #                 ||||||||||||||||||||||||||||| || ||   (next state)
        #                 |||||||||||||||||||||||||||||+||+||-- Group $4
        if ($cur_line =~ /\s*(\d+):\s*(\w+)\s*(.*)\s*\((.*)\)/) {
            push(@re_debug, {
                    node => $1,
                    type => $2,
                    raw_type => $2,
                    arg => $3,
                    next => $4
                    });
            next;
        }
        if ($cur_line =~ /^anchored/) {
            next;
        }
        if ($cur_line =~ /^Freeing/) {
            last;
        }
    }
}
########################################################
# $new_index = parse_node($index, $array, $next, $close)
#       -- Parse a single regular expression node
#       -- Stop when next (or end) is found
#       -- Or when a close ")" is found
########################################################
sub parse_node($$$$);
sub parse_node($$$$)
{
    my $index = shift;          # Index into the array
    my $array = shift;          # Array to put things on
    my $next = shift;           # Next node
    my $close = shift;          # Looking for a close?

    my $min_flag = 0;           # Minimize flag
    while (1) {
        if (not defined($re_debug[$index])) {
            return ($index);
        }
        if (defined($next)) {
            if ($next <= $re_debug[$index]->{node}) {
                return ($index);
            }
        }
        if ($re_debug[$index]->{type} =~ /CLOSE(\d+)/) {
            if (defined($close)) {
                if ($1 == $close) {
                    return ($index + 1);
                }
            }
        }
        if ($re_debug[$index]->{type} eq "MINMOD") {
            $min_flag = 1;
            $index++;
            next;
        }
#------------------------------------------------------------
        if (($re_debug[$index]->{type} eq "IFMATCH") ||
            ($re_debug[$index]->{type} eq "UNLESSM")) {
            if ($re_debug[$index]->{arg} !~ /\[(.*?)\]/) {
                die("IFMATCH/UNLESSM funny argument $re_debug[$index]->{arg}");
            }
            my $equal = "!=";   # Ending text (= or !=)
            my $color = "Red";  # Color for box
            if ($re_debug[$index]->{type} eq "IFMATCH") {
                $equal = "=";
                $color = "Green";
            }
            my $flag = $1;      # Flag indicating the next look ahead
            my $text;           # Text to label this box
            if ($flag eq "-0") {
                $text = "$equal ahead";
            } elsif ($flag eq "-0") {
                $text = "$equal behind";
            } elsif ($flag eq "-1") {
                $text = "$equal behind";
            } else {
                die("Unknown IFMATCH/UNLESSM flag text $flag");
                exit;
            }
            push(@{$array}, {node => $re_debug[$index],
                            text => $text,
                            color => $color,
                           children => []});

            $index = parse_node($index+1, $$array[$#$array]->{children},
                    $re_debug[$index]->{next}, undef);
            next;
        }
#------------------------------------------------------------
        if ($re_debug[$index]->{type} =~ /OPEN(\d+)/) {
            my $paren_count = $1;
            $re_debug[$index]->{type} = "OPEN";
            push(@{$array}, {node => $re_debug[$index],
                            paren_count => $paren_count,
                            text => "( ) => \$$paren_count",
                            color => "Brown",
                           children => []});

            $index = parse_node($index+1, $$array[$#$array]->{children},
                    undef, $paren_count);
            next;
        }
#------------------------------------------------------------
        if ($re_debug[$index]->{type} =~ /REF(\d+)/) {
            my $ref_number = $1;
            $re_debug[$index]->{type} = "REF";
            push(@{$array}, {node => $re_debug[$index],
                            ref => $ref_number,
                           children => []});

            ++$index;
            next;
        }
#------------------------------------------------------------
        if ($re_debug[$index]->{type} eq "BRANCH") {
            push(@{$array}, {node => $re_debug[$index],
                           choices => []});
            my $choice_index = 0;
            while (1) {
                # Next node in this series
                my $next = $re_debug[$index]->{next};

                $$array[$#$array]->{choices}[$choice_index] = [];

                $index = parse_node($index+1,
                        $$array[$#$array]->{choices}[$choice_index],
                        $next, undef);

                if (not defined($re_debug[$index])) {
                    last;
                }

                if ($re_debug[$index]->{type} ne "BRANCH") {
                    last;
                }
                $choice_index++;
            }
            next;
        }
#------------------------------------------------------------
        if ($re_debug[$index]->{type} eq "CURLYX") {
            my $min_number;     # Min number of matches
            my $max_number;     # Max number of matches

            if ($re_debug[$index]->{arg} =~ /{(\d+),(\d+)}/) {
                $min_number = $1;
                $max_number = $2;
            } else {
                die("Funny CURLYX args $re_debug[$index]->{arg}");
                exit;
            }
            if ((($min_number == 0)||($min_number == 1)) && ($max_number == 32767)) {
                # Are we doing a * or + (anything else is just too hard_
                my $star_flag = ($min_number == 0);

                # Node that's enclosed inside this one
                my $child = {
                        node => {
                                    type => ($star_flag) ? "STAR" : "PLUS",
                                    raw_type => $re_debug[$index]->{type},
                                    arg => $re_debug[$index]->{arg},
                                    next => $re_debug[$index]->{next}
                                },
                                min_flag => $min_flag,
                                children => [],
                        };

                push(@{$array}, $child);

                $index = parse_node($index+1, $child->{children},
                        $re_debug[$index]->{next}, undef);
                next;
            } else {
                die("Unexpected CURLYX count ($re_debug[$index]->{arg})");
                exit;
            }
        }
#------------------------------------------------------------
        if ($re_debug[$index]->{type} eq "CURLYM") {

            my $paren_count;    # () number 
            my $min_number;     # Min number of matches
            my $max_number;     # Max number of matches

            if ($re_debug[$index]->{arg} =~ /\[(\d+)\]\s*{(\d+),(\d+)}/) {
                $paren_count = $1;
                $min_number = $2;
                $max_number = $3;
            } else {
                die("Funny CURLYM args $re_debug[$index]->{arg}");
                exit;
            }
            if ((($min_number == 0)||($min_number == 1)) && ($max_number == 32767)) {
                # Are we doing a * or + (anything else is just too hard_
                my $star_flag = ($min_number == 0);

                # Node that's enclosed inside this one
                my $child = {
                        node => {
                                    type => ($star_flag) ? "STAR" : "PLUS",
                                    raw_type => $re_debug[$index]->{type},
                                    arg => $re_debug[$index]->{arg},
                                    next => $re_debug[$index]->{next}
                                },
                                min_flag => $min_flag,
                                children => [],
                        };
                $min_flag = 0;

                # The text for labeling this node
                my $text = "( ) => \$$paren_count";
                if ($paren_count == 0) {
                    $text = '( ) [no $x]';
                }
                push(@{$array},
                        {node => {
                                    type => "OPEN",
                                    raw_type => $re_debug[$index]->{type},
                                    arg => $re_debug[$index]->{arg},
                                    next => $re_debug[$index]->{next}},
                                    color => "Brown",
                         paren_count => $paren_count,
                         text => $text,
                         children => [$child]});

                $index = parse_node($index+1, $child->{children},
                        $re_debug[$index]->{next}, undef);
                next;
            } else {
                die("Unexpected CURLY count ($re_debug[$index]->{arg})");
                exit;
            }
        }
#------------------------------------------------------------
        if ($re_debug[$index]->{type} eq "STAR") {
            push(@{$array}, {node => $re_debug[$index],
                           min_flag => $min_flag,
                           children => []});
            $min_flag = 0;

            my $star_next;      # Where we go for the next state
            if (defined($next)) {
                $star_next = $next;
            } else {
                $star_next = $re_debug[$index]->{next};
            }

            $index = parse_node($index+1, $$array[$#$array]->{children},
                    $star_next, undef);
            next;
        }
#------------------------------------------------------------
        if ($re_debug[$index]->{type} eq "PLUS") {
            push(@{$array}, {node => $re_debug[$index],
                           min_flag => $min_flag,
                           children => []});
            $min_flag = 0;
            $index = parse_node($index+1, $$array[$#$array]->{children},
                    $re_debug[$index]->{next}, undef);
            next;
        }
#------------------------------------------------------------
        # Ignore a couple of nodes
        if ($re_debug[$index]->{type} eq "WHILEM") {
            ++$index;
            next;
        }
        if ($re_debug[$index]->{type} eq "SUCCEED") {
            ++$index;
            next;
        }
        if ($re_debug[$index]->{type} eq "NOTHING") {
            ++$index;
            next;
        }
        if ($re_debug[$index]->{type} eq "TAIL") {
            ++$index;
            next;
        }
        push(@$array, {node => $re_debug[$index]});
        if ($re_debug[$index]->{type} eq "END") {
            return ($index+1);
        }
        $index++;

    }
}
########################################################
# do_size($cur_node) -- Compute the size of a given node
########################################################
sub do_size($);
sub do_size($)
{
    my $cur_node = shift;

    if (not defined($compute_size{$cur_node->{node}->{type}})) {
        die("No compute function for $cur_node->{node}->{type}");
        exit;
    }
    $compute_size{$cur_node->{node}->{type}}($cur_node);
}
########################################################
# size_array(\@array) -- Compute the size of an array of nodes
#
# Returns
#       (x_size, y_size) -- Size of the elements
#
#       x_size -- Size of all the elements in X
#               (We assume they are laid out in a line)
#       y_size -- Biggest Y size (side by side layout)
########################################################
sub size_array(\@)
{
    my $re_array = shift;       # The array
    my $x_size = 0;             # Size of the array in X
    my $y_size = 0;             # Size of the elements in Y

    foreach my $cur_node(@$re_array) {
        do_size($cur_node);
        $x_size += $cur_node->{x_size};
        if ($y_size < $cur_node->{y_size}) {
            $y_size = $cur_node->{y_size};
        }
    }
    return ($x_size, $y_size);
}
########################################################
# layout_array($x_start, $y_start, $y_max, \@array)
#
# Layout an array of nodes
########################################################
sub layout_array($$$\@)
{
    my $x_start = shift;        # Starting point in X
    my $y_start = shift;        # Starting point in Y
    my $y_max = shift;          # largest Y value
    my $re_array = shift;       # The data

    foreach my $cur_node(@$re_array) {
        $cur_node->{x_loc} = $x_start;
        $cur_node->{y_loc} = $y_start + ($y_max - $cur_node->{y_size})/2;
        $x_start += $cur_node->{x_size};
    }
}
########################################################
# convert_re -- Convert @re_debug -> @format_re
#
# The formatted re node contains layout information as
# well as information on nodes contained inside the current
# one.
########################################################
sub convert_re()
{
    # Clear out old data
    @format_re = ();

    parse_node(0, \@format_re, undef, undef);
    #
    # Compute sizes of each node
    #
    my ($x_size, $y_size) = size_array(@format_re);
    #
    # Compute the location of each node
    #
    layout_array(100, 100, $y_size, @format_re);
}


##############################################################
# draw_node_array -- draw an array of nodes
##############################################################
sub draw_node_array($)
{
    my $array = shift;
    #
    # Draw Nodes
    #
    foreach my $cur_node (@$array) {
        if (not defined($draw_node{$cur_node->{node}->{type}})) {
            die("No draw function for $cur_node->{node}->{type}");
        }
        $draw_node{$cur_node->{node}->{type}}($cur_node);
    }
    #
    # Loop through all the things (except the last) and
    # draw arrows between them
    #
    for (my $index = 0; $index < $#$array; ++$index) {
        my $from_x = $array->[$index]->{right_x};
        my $from_y = $array->[$index]->{right_y};

        my $to_x = $array->[$index+1]->{left_x};
        my $to_y = $array->[$index+1]->{left_y};

        $tk_canvas->createLine(
                $from_x, $from_y,
                $to_x, $to_y,
                -arrow => "last",
                -tags => "arrow",
                -width => 3,
                -fill => "Black"
        );
    }
}
##############################################################
# redraw_canvas -- Draw the canvas
##############################################################
sub redraw_canvas()
{
    if (not defined($tk_canvas)) {
        return; # Called too early
    }
    $tk_canvas->delete("all");
    ############################################################
    # Determine the location of each of the nodes
    ############################################################

    draw_node_array(\@format_re);
    # Make all the canvas visible
    my @bbox = $tk_canvas->bbox('all');
    if ($#bbox >= 0)
    {
	$tk_canvas->configure(-scrollregion => \@bbox);
    }
}

#################################################################
# do_entry(parent, label, value) -- Create an text entry blank
#################################################################
sub do_entry($$$)
{
    my $tk_parent = shift;      # Parent window
    my $label = shift;          # Label for the item
    my $var = shift;            # Variable to change

    my $tk_line = $tk_parent->Frame(
        )->pack(
            -side => "top",
            -expand => 1,
            -fill => 'x'
        );
    $tk_line->Label(
                -text => $label
            )->pack(
                -side => 'left'
            );
    $tk_line->Entry(
                -textvariable => $var,
                -width => 3
            )->pack(
                -side => 'left'
            );
}
################################################################
# cd_print -- Handle the print button.
#
# Warning: Does not work well
################################################################
sub cb_print()
{
    $tk_canvas->postscript(-file => "tmp.ps");
    print "Done\n";
}
################################################################
# new_re -- Called when the re changes
################################################################
sub new_re($)
{
    my $inc = shift;    # Increment to use

    # Index of expression we have now
    my $index = $tk_list->curselection();
    $tk_list->selectionClear($index);

    $index += $inc;

    if ($index < 0) {
        $index = 0;
    }

    # Refresh the list of regular expressions
    @re_list = $tk_list->get(0, 'end');

    if ($index > $#re_list) {
        $index = $#re_list;
    }
    $current_re = $re_list[$index];
    $tk_current_re->delete("0.0", "end");
    $tk_current_re->insert("0.0", $current_re);

    $tk_list->selectionSet($index);

    parse_re();
    convert_re();
    redraw_canvas();
}
################################################################
# add_re -- Add the re to the list
################################################################
sub add_re()
{
    $tk_list->insert("active", $re_to_add);
    new_re(1);
}
################################################################
# Create (or popup) the window containing the RE's we're using
################################################################
sub make_re_list()
{
    if (defined($tk_re_list_window)) {
	$tk_re_list_window->deiconify();
	$tk_re_list_window->raise();
        return;
    }
    $tk_re_list_window = $tk_mw->Toplevel(-background=>"white");
    $tk_re_list_window->title('Regular Expression List');
    $tk_re_list_window->iconname('Listbox');
    $tk_re_list_window->minsize(30, 10);

    $tk_re_list_window->Label(
            -justify => 'left',
            -text => 'Regular Expression List',
            -background => "white"
    )->pack(
        -side => "top"
    );

    $tk_list  = $tk_re_list_window->ScrlListbox(
            -scrollbars => 'osow',
            -background => "white",
            -selectmode => "Single"
        )->pack(
            -fill => 'both',
            -expand => "yes"
        );

    my $button_frame = $tk_re_list_window->Frame(
                -background => "white"
            )->pack(
                -side => "top",
                -fill => 'x',
                -expand => "yes"
            );

    $button_frame->Entry(
            -background => "white",
            -width => 30,
            -textvariable => \$re_to_add
        )->pack(
            -side => "top",
            -fill => "x",
            -expand => "yes"
        );

    $button_frame->Button(
            -text => "OK",
            -background => "LightGreen",
            -command => sub {new_re(0); $tk_re_list_window->withdraw();}
        )->pack(
            -side => "left"
        );
    $button_frame->Button(
            -text => "New",
            -background => "LightBlue",
            -command => \&add_re
        )->pack(
            -side => "right"
        );

    $tk_list->insert(0, @re_list);
    $tk_list->bind('<Double-1>', sub {new_re(0);} );
    $tk_re_list_window->withdraw();
} 

################################################################
# build_gui -- Create the gui
################################################################
sub build_gui()
{
    # The top level window
    $tk_mw = MainWindow->new();
    $tk_mw->title("Regular Expression Graph");

    # A place to put buttons
    my $tk_button_frame = $tk_mw->Frame(
                -background => "gray"
	    )->pack(
		-side => 'top',
		-fill => 'x'
	    );

    if ($opt_P) {
        $tk_button_frame->Button(
                -text => "Print",
                -command => \&cb_print,
                -background => "lightblue"
            )->pack(
                -side => 'left'
            );
    }

    $tk_button_frame->Button(
            -text => "Next",
            -background => "lightblue",
            -command => sub { new_re(+1); }
        )->pack(
            -side => 'left'
        );

    $tk_button_frame->Button(
            -text => "Previous",
            -background => "lightblue",
            -command => sub { new_re(-1); }
        )->pack(
            -side => 'left'
        );

    $tk_current_re = $tk_button_frame->ROText(
            -width => 30,
            -height => 1,
            -background => "white"
        )->pack(
            -side => 'left'
        );
    $tk_current_re->insert("0.0", $current_re);

    $tk_button_frame->Button(
            -text => "List",
            -background => "lightblue",
            -command => \&make_re_list
        )->pack(
            -side => 'left'
        );

    # The exit button
    $tk_button_frame->Button(
	    -text => "Exit",
	    -command => sub {exit(0);}
	)->pack(
	    -side => 'right'
	);

    # The canvas as scrolled
    $tk_scrolled_canvas = $tk_mw->Scrolled(
		'Canvas',
		-background => 'white',
		-width => 1000,
		-height => 600,
		-scrollbars => "osoe"
	    )->pack(
		-side => 'bottom',
		-fill => 'both',
		-expand => 1
	    );
    # The canvas itself
    $tk_canvas = $tk_scrolled_canvas->Subwidget('canvas');

    # Adapted from the Tk/demos/mkScroll.pl file that comes with Tk
    $tk_canvas->Tk::bind('<2>' => sub {
	    my ($c) = @_;
	    my $e = $c->XEvent;
	    $c->scan('mark', $e->x, $e->y);
	});
    $tk_canvas->Tk::bind('<3>' => sub {
	    my ($c) = @_;
	    my $e = $c->XEvent;
	    $c->scan('mark', $e->x, $e->y);
	});
    $tk_canvas->Tk::bind('<B2-Motion>' => sub {
	    my ($c) = @_;
	    my $e = $c->XEvent;
	    $c->scan('dragto', $e->x, $e->y);
	});
    $tk_canvas->Tk::bind('<B3-Motion>' => sub {
	    my ($c) = @_;
	    my $e = $c->XEvent;
	    $c->scan('dragto', $e->x, $e->y);
	});

}

# -f:file -- File contains the RE
# -P    -- Turn on print button
my $status = getopts("df:P");
if ($status == 0)
{
    usage();
}

@re_list = @ARGV;
if ($opt_f) {
    my @expr = `cat $opt_f`;
    chomp(@expr);
    push(@re_list, @expr);
}
if ($#re_list == -1) {
    @re_list = (
        'test',
        '^Test',
        '^ *#',
        '^[ \\t]*#',
        '^\\s*#',
        '([^#]*)(#.*)',
        'a|b',
        '^(([^#"]*|".*")*)(#.*)',
        '^((?:[^#"]*|".*?")*)(#.*)',
        '^((?:[^#"]*|".*?(?<!\\\\)")*)(#.*)'
    );
}

build_gui();
make_re_list();
$current_re = $re_list[0];
$tk_current_re->insert("0.0", $current_re);
$tk_list->selectionSet(0);
parse_re();
convert_re();

$tk_font = $tk_canvas->Font(
    -family => "courier",
    -size => "24",
    -weight => "bold");

$tk_small_font = $tk_canvas->Font(
    -family => "courier",
    -size => "18",
    -weight => "bold");

redraw_canvas();

MainLoop();

__END__

		    GNU GENERAL PUBLIC LICENSE
		       Version 2, June 1991

 Copyright (C) 1989, 1991 Free Software Foundation, Inc.
     59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.

			    Preamble

  The licenses for most software are designed to take away your
freedom to share and change it.  By contrast, the GNU General Public
License is intended to guarantee your freedom to share and change free
software--to make sure the software is free for all its users.  This
General Public License applies to most of the Free Software
Foundation's software and to any other program whose authors commit to
using it.  (Some other Free Software Foundation software is covered by
the GNU Library General Public License instead.)  You can apply it to
your programs, too.

  When we speak of free software, we are referring to freedom, not
price.  Our General Public Licenses are designed to make sure that you
have the freedom to distribute copies of free software (and charge for
this service if you wish), that you receive source code or can get it
if you want it, that you can change the software or use pieces of it
in new free programs; and that you know you can do these things.

  To protect your rights, we need to make restrictions that forbid
anyone to deny you these rights or to ask you to surrender the rights.
These restrictions translate to certain responsibilities for you if you
distribute copies of the software, or if you modify it.

  For example, if you distribute copies of such a program, whether
gratis or for a fee, you must give the recipients all the rights that
you have.  You must make sure that they, too, receive or can get the
source code.  And you must show them these terms so they know their
rights.

  We protect your rights with two steps: (1) copyright the software, and
(2) offer you this license which gives you legal permission to copy,
distribute and/or modify the software.

  Also, for each author's protection and ours, we want to make certain
that everyone understands that there is no warranty for this free
software.  If the software is modified by someone else and passed on, we
want its recipients to know that what they have is not the original, so
that any problems introduced by others will not reflect on the original
authors' reputations.

  Finally, any free program is threatened constantly by software
patents.  We wish to avoid the danger that redistributors of a free
program will individually obtain patent licenses, in effect making the
program proprietary.  To prevent this, we have made it clear that any
patent must be licensed for everyone's free use or not licensed at all.

  The precise terms and conditions for copying, distribution and
modification follow.

		    GNU GENERAL PUBLIC LICENSE
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  0. This License applies to any program or other work which contains
a notice placed by the copyright holder saying it may be distributed
under the terms of this General Public License.  The "Program", below,
refers to any such program or work, and a "work based on the Program"
means either the Program or any derivative work under copyright law:
that is to say, a work containing the Program or a portion of it,
either verbatim or with modifications and/or translated into another
language.  (Hereinafter, translation is included without limitation in
the term "modification".)  Each licensee is addressed as "you".

Activities other than copying, distribution and modification are not
covered by this License; they are outside its scope.  The act of
running the Program is not restricted, and the output from the Program
is covered only if its contents constitute a work based on the
Program (independent of having been made by running the Program).
Whether that is true depends on what the Program does.

  1. You may copy and distribute verbatim copies of the Program's
source code as you receive it, in any medium, provided that you
conspicuously and appropriately publish on each copy an appropriate
copyright notice and disclaimer of warranty; keep intact all the
notices that refer to this License and to the absence of any warranty;
and give any other recipients of the Program a copy of this License
along with the Program.

You may charge a fee for the physical act of transferring a copy, and
you may at your option offer warranty protection in exchange for a fee.

  2. You may modify your copy or copies of the Program or any portion
of it, thus forming a work based on the Program, and copy and
distribute such modifications or work under the terms of Section 1
above, provided that you also meet all of these conditions:

    a) You must cause the modified files to carry prominent notices
    stating that you changed the files and the date of any change.

    b) You must cause any work that you distribute or publish, that in
    whole or in part contains or is derived from the Program or any
    part thereof, to be licensed as a whole at no charge to all third
    parties under the terms of this License.

    c) If the modified program normally reads commands interactively
    when run, you must cause it, when started running for such
    interactive use in the most ordinary way, to print or display an
    announcement including an appropriate copyright notice and a
    notice that there is no warranty (or else, saying that you provide
    a warranty) and that users may redistribute the program under
    these conditions, and telling the user how to view a copy of this
    License.  (Exception: if the Program itself is interactive but
    does not normally print such an announcement, your work based on
    the Program is not required to print an announcement.)

These requirements apply to the modified work as a whole.  If
identifiable sections of that work are not derived from the Program,
and can be reasonably considered independent and separate works in
themselves, then this License, and its terms, do not apply to those
sections when you distribute them as separate works.  But when you
distribute the same sections as part of a whole which is a work based
on the Program, the distribution of the whole must be on the terms of
this License, whose permissions for other licensees extend to the
entire whole, and thus to each and every part regardless of who wrote it.

Thus, it is not the intent of this section to claim rights or contest
your rights to work written entirely by you; rather, the intent is to
exercise the right to control the distribution of derivative or
collective works based on the Program.

In addition, mere aggregation of another work not based on the Program
with the Program (or with a work based on the Program) on a volume of
a storage or distribution medium does not bring the other work under
the scope of this License.

  3. You may copy and distribute the Program (or a work based on it,
under Section 2) in object code or executable form under the terms of
Sections 1 and 2 above provided that you also do one of the following:

    a) Accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of Sections
    1 and 2 above on a medium customarily used for software interchange; or,

    b) Accompany it with a written offer, valid for at least three
    years, to give any third party, for a charge no more than your
    cost of physically performing source distribution, a complete
    machine-readable copy of the corresponding source code, to be
    distributed under the terms of Sections 1 and 2 above on a medium
    customarily used for software interchange; or,

    c) Accompany it with the information you received as to the offer
    to distribute corresponding source code.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form with such
    an offer, in accord with Subsection b above.)

The source code for a work means the preferred form of the work for
making modifications to it.  For an executable work, complete source
code means all the source code for all modules it contains, plus any
associated interface definition files, plus the scripts used to
control compilation and installation of the executable.  However, as a
special exception, the source code distributed need not include
anything that is normally distributed (in either source or binary
form) with the major components (compiler, kernel, and so on) of the
operating system on which the executable runs, unless that component
itself accompanies the executable.

If distribution of executable or object code is made by offering
access to copy from a designated place, then offering equivalent
access to copy the source code from the same place counts as
distribution of the source code, even though third parties are not
compelled to copy the source along with the object code.

  4. You may not copy, modify, sublicense, or distribute the Program
except as expressly provided under this License.  Any attempt
otherwise to copy, modify, sublicense or distribute the Program is
void, and will automatically terminate your rights under this License.
However, parties who have received copies, or rights, from you under
this License will not have their licenses terminated so long as such
parties remain in full compliance.

  5. You are not required to accept this License, since you have not
signed it.  However, nothing else grants you permission to modify or
distribute the Program or its derivative works.  These actions are
prohibited by law if you do not accept this License.  Therefore, by
modifying or distributing the Program (or any work based on the
Program), you indicate your acceptance of this License to do so, and
all its terms and conditions for copying, distributing or modifying
the Program or works based on it.

  6. Each time you redistribute the Program (or any work based on the
Program), the recipient automatically receives a license from the
original licensor to copy, distribute or modify the Program subject to
these terms and conditions.  You may not impose any further
restrictions on the recipients' exercise of the rights granted herein.
You are not responsible for enforcing compliance by third parties to
this License.

  7. If, as a consequence of a court judgment or allegation of patent
infringement or for any other reason (not limited to patent issues),
conditions are imposed on you (whether by court order, agreement or
otherwise) that contradict the conditions of this License, they do not
excuse you from the conditions of this License.  If you cannot
distribute so as to satisfy simultaneously your obligations under this
License and any other pertinent obligations, then as a consequence you
may not distribute the Program at all.  For example, if a patent
license would not permit royalty-free redistribution of the Program by
all those who receive copies directly or indirectly through you, then
the only way you could satisfy both it and this License would be to
refrain entirely from distribution of the Program.

If any portion of this section is held invalid or unenforceable under
any particular circumstance, the balance of the section is intended to
apply and the section as a whole is intended to apply in other
circumstances.

It is not the purpose of this section to induce you to infringe any
patents or other property right claims or to contest validity of any
such claims; this section has the sole purpose of protecting the
integrity of the free software distribution system, which is
implemented by public license practices.  Many people have made
generous contributions to the wide range of software distributed
through that system in reliance on consistent application of that
system; it is up to the author/donor to decide if he or she is willing
to distribute software through any other system and a licensee cannot
impose that choice.

This section is intended to make thoroughly clear what is believed to
be a consequence of the rest of this License.

  8. If the distribution and/or use of the Program is restricted in
certain countries either by patents or by copyrighted interfaces, the
original copyright holder who places the Program under this License
may add an explicit geographical distribution limitation excluding
those countries, so that distribution is permitted only in or among
countries not thus excluded.  In such case, this License incorporates
the limitation as if written in the body of this License.

  9. The Free Software Foundation may publish revised and/or new versions
of the General Public License from time to time.  Such new versions will
be similar in spirit to the present version, but may differ in detail to
address new problems or concerns.

Each version is given a distinguishing version number.  If the Program
specifies a version number of this License which applies to it and "any
later version", you have the option of following the terms and conditions
either of that version or of any later version published by the Free
Software Foundation.  If the Program does not specify a version number of
this License, you may choose any version ever published by the Free Software
Foundation.

  10. If you wish to incorporate parts of the Program into other free
programs whose distribution conditions are different, write to the author
to ask for permission.  For software which is copyrighted by the Free
Software Foundation, write to the Free Software Foundation; we sometimes
make exceptions for this.  Our decision will be guided by the two goals
of preserving the free status of all derivatives of our free software and
of promoting the sharing and reuse of software generally.

			    NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.

		     END OF TERMS AND CONDITIONS

	    How to Apply These Terms to Your New Programs

  If you develop a new program, and you want it to be of the greatest
possible use to the public, the best way to achieve this is to make it
free software which everyone can redistribute and change under these terms.

  To do so, attach the following notices to the program.  It is safest
to attach them to the start of each source file to most effectively
convey the exclusion of warranty; and each file should have at least
the "copyright" line and a pointer to where the full notice is found.

    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) <year>  <name of author>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


Also add information on how to contact you by electronic and paper mail.

If the program is interactive, make it output a short notice like this
when it starts in an interactive mode:

    Gnomovision version 69, Copyright (C) year  name of author
    Gnomovision comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
    This is free software, and you are welcome to redistribute it
    under certain conditions; type `show c' for details.

The hypothetical commands `show w' and `show c' should show the appropriate
parts of the General Public License.  Of course, the commands you use may
be called something other than `show w' and `show c'; they could even be
mouse-clicks or menu items--whatever suits your program.

You should also get your employer (if you work as a programmer) or your
school, if any, to sign a "copyright disclaimer" for the program, if
necessary.  Here is a sample; alter the names:

  Yoyodyne, Inc., hereby disclaims all copyright interest in the program
  `Gnomovision' (which makes passes at compilers) written by James Hacker.

  <signature of Ty Coon>, 1 April 1989
  Ty Coon, President of Vice

This General Public License does not permit incorporating your program into
proprietary programs.  If your program is a subroutine library, you may
consider it more useful to permit linking proprietary applications with the
library.  If this is what you want to do, use the GNU Library General
Public License instead of this License.
